-----------------------------------------------------------------------------
-- |
-- Module      :  CPLSystem
-- Copyright   :  (c) Masahiro Sakai 2004
-- License     :  BSD-style
-- 
-- Maintainer  :  sakai@tom.sfc.keio.ac.jp
-- Stability   :  provisional
-- Portability :  portable
--
-----------------------------------------------------------------------------

module CPLSystem
    ( VarTable
    , System (..)
    , emptySystem
    , parseExp
    , checkName
    , parseCDT
    , addCDT
    , letExp
    , simp
    ) where

import CDT
import qualified CDTParser
import qualified ExpParser
import Exp
import qualified AExp
import qualified Simp
import qualified Typing
import Typing (Typing(..))

import Maybe
import List
import Text.ParserCombinators.Parsec
import qualified Data.Map as Map

type CDTEnv = [CDT]

type VarTable = Map.Map String (Simp.CompiledExp,AExp.TypeScheme)
data System
    = System
    { objects  :: !CDTEnv
    , varTable :: !VarTable
    , trace    :: !Bool
    , lastExp  :: !(Maybe Exp)
    }

emptySystem :: System
emptySystem =
    System{ objects  = []
          , varTable = Map.empty
          , trace    = False
          , lastExp  = Nothing
          }

parseExp :: Monad m => System -> String -> m (Int, Typing)
parseExp sys str =
    case parse ExpParser.exp "" str of
    Left err -> fail (show err)
    Right e ->
        case ExpParser.evalExp (objects sys) checkVar e of
        Nothing -> fail "invalid expression" -- FIXME
        Just e1 ->
            Typing.runTI typEnv $
                do t <- Typing.inferType e2
                   t <- Typing.appSubst t
                   return $ AExp.quantify t
            where e2 = case lastExp sys of
                       Just it -> substIt it e1
                       Nothing -> e1
    where checkVar name = isJust (Map.lookup name (varTable sys))
                          || (isJust (lastExp sys) && name == "it")
          typEnv = Map.map snd (varTable sys)

-- Exp.hs に移動?
substIt :: Exp -> Exp -> Exp
substIt it = h
      where h Identity         = Identity
            h (Comp a b)       = Comp (h a) (h b)
            h e@(Nat _)        = e
            h (Fact obj args)  = Fact obj (map h args)
            h (Funct obj args) = Funct obj (map h args)
            h (Var "it")       = it
            h e@(Var _)        = e

checkName :: Monad m => System -> String -> m ()
checkName sys name =
    let vt    = varTable sys
        objs  = objects sys
        names = Map.keys vt ++
                map CDT.functName objs ++
                map CDT.factName objs  ++
                concatMap (map CDT.natName . CDT.nats) objs
    in if name `elem` names
           then fail ("\"" ++ name ++ "\" is already used")
           else return ()

parseCDT :: Monad m => System -> String -> m CDT.CDT
parseCDT sys src = case parse CDTParser.cdtDecl "" src of
                   Left err   -> fail (show err)
                   Right decl -> CDTParser.evalCDTDecl (objects sys) decl

addCDT :: Monad m => System -> CDT.CDT -> m System
addCDT sys obj =
    if CDT.isComputable obj
    then do checkName sys (CDT.functName obj)
            checkName sys (CDT.factName obj)
            mapM_ (checkName sys . CDT.natName) (CDT.nats obj)
            return sys{ objects = obj : objects sys }
    else fail "not a computable object"

compile :: System -> Exp -> Simp.CompiledExp
compile sys e = Simp.compile varLookup e
  where
    varLookup name = fromJust $ do
      (exp,_) <- Map.lookup name (varTable sys)
      return exp

letExp :: Monad m => System -> String -> (Int, Typing) -> m System
letExp sys name (n, e :! t) = do
  checkName sys name
  return sys{ varTable = Map.insert name entry (varTable sys) }
  where
    entry = (compile sys (AExp.skelton e), (n, t))

simp :: System -> Bool -> Exp -> [(Int, Simp.CompiledExp, Simp.CompiledExp)]
simp sys full exp =
    let exp' = compile sys exp
        traces =
            if trace sys
               then Simp.simpWithTrace full exp'
               else [(0, compile sys Identity, Simp.simp full exp')]
    in traces
