module Typing
    ( Typing (..)
    , inferType

    , TI
    , Env
    , runTI
    , getSubst
    , extSubst
    , appSubst
    , unify
    , newFEVar
    ) where

import Variance
import Funct
import qualified FE
import qualified CDT
import Type (GenType(..), Type)
import Subst
import AExp
import qualified Exp as E

import Data.List (nub)
import Control.Monad.RWS
import qualified Data.Map as Map

----------------------------------------------------------------------------

infix 6 :!
data Typing = AExp :! Type

tm :: Typing -> AExp
tm (e :! _) = e

ty :: Typing -> Type
ty (_ :! t) = t

instance FEs Typing where
    tv (e :! t)      = nub (tv e ++ tv t)
    apply s (e :! t) = apply s e :! apply s t

----------------------------------------------------------------------------

-- Type inference monad transformer
type TI m = RWST Env () (Int, Subst) m

type Env = Map.Map E.Id TypeScheme

runTI :: Monad m => Env -> TI m a -> m a
runTI env m = liftM fst (evalRWST m env initialState)
    where initialState = (0, nullSubst)

getSubst :: Monad m => TI m Subst
getSubst = do (_,s) <- get
              return s

extSubst :: Monad m => Subst -> TI m ()
extSubst s2 =
    do (i,s1) <- get
       put (i, s2@@s1)
       return ()

appSubst :: (Monad m, FEs a) => a -> TI m a
appSubst x =
    do s <- getSubst
       return (apply s x)

unify :: Monad m => FE.FE -> FE.FE -> TI m ()
unify a b =
    do s1 <- getSubst
       s2 <- mgu (apply s1 a) (apply s1 b)
       extSubst s2

newFEVar :: Monad m => TI m FE.FE
newFEVar = do (i,s) <- get
              put (i+1,s)
              return $ FE.Var i

----------------------------------------------------------------------------

inferType :: Monad m => E.Exp -> TI m Typing
inferType E.Identity = iIdentity
inferType (E.Comp a b) = 
    do a' <- inferType a
       b' <- inferType b
       iComp a' b'
inferType (E.Nat nat)        = iNat nat
inferType (E.Fact obj args)  = iFact obj  =<< mapM inferType args
inferType (E.Funct obj args) = iFunct obj =<< mapM inferType args
inferType (E.Var v)          = iVar v

----------------------------------------------------------------------------
-- introduction rules

iIdentity :: Monad m => TI m Typing
iIdentity = 
    do v <- newFEVar
       return $ Identity v :! v :-> v

iComp :: Monad m => Typing -> Typing -> TI m Typing
iComp (f :! domf :-> codf) (g :! domg :-> codg) =
    do unify domf codg
       return $ Comp f g :! domg :-> codf

iNat :: Monad m => CDT.Nat -> TI m Typing
iNat nat =
    do annotation <- sequence $ replicate (CDT.natNTypeParams nat) newFEVar
       return $ Nat nat annotation :!
                apply (zip [0..] annotation) (CDT.natType nat)

iFact :: Monad m => CDT.CDT -> [Typing] -> TI m Typing
iFact obj args =
    do annotation <- sequence $ replicate (CDT.factNTypeParams obj) newFEVar
       let s = zip [0..] annotation
       unifyParams (apply s (CDT.factParams obj)) args
       return $ Fact obj (map tm args) annotation :!
                apply s (CDT.factDestType obj)
    where unifyParams [] [] = return ()
          unifyParams ((pdom :-> pcod):ps) ((_ :! adom:->acod):as) =
              do unify pdom adom
                 unify pcod acod
                 unifyParams ps as
                 return ()
          unifyParams _ _ = fail "wrong number of arguments"

iFunct :: Monad m => CDT.CDT -> [Typing] -> TI m Typing
iFunct obj args =
    do (doml,codl) <- unifyParams (variance obj) args
       return $ Funct obj (map tm args) :! FE.Ap obj doml :-> FE.Ap obj codl
    where unifyParams [] [] = return ([],[])
          unifyParams (v:vs) ((_ :! d:->c) : as) = 
              do (doms,cods) <- unifyParams vs as
                 case v of
                     Covariance     -> return (d:doms, c:cods)
                     Contravariance -> return (c:doms, d:cods)
                     FixedVariance  -> do unify d c
                                          return (d:doms, d:cods)
                     FreeVariance   -> return (d:doms, c:cods)
          unifyParams _ _ = fail "wrong number of arguments"

iVar :: Monad m => E.Id -> TI m Typing
iVar v =
    do env <- ask
       case Map.lookup v env of
           Just (n,typ) ->
               do annotation <- sequence (replicate n newFEVar)
                  let t = apply (zip [0..] annotation) typ
                  return $ Var v annotation t :! t
           Nothing -> fail $ "no such variable: " ++ v

----------------------------------------------------------------------------
